@@ -15,11 +15,18 @@
 #include "clang/AST/StmtVisitor.h"
 #include "clang/Sema/Sema.h"
 
+#include <llvm/ADT/ArrayRef.h>
+
 #include <array>
+#include <limits>
 #include <memory>
 #include <stack>
 #include <unordered_map>
 
+namespace llvm {
+template <typename T> class SmallVectorImpl;
+}
+
 namespace clad {
   class ErrorEstimationHandler;
   class ExternalRMVSource;
@@ -30,8 +37,7 @@ namespace clad {
   class ReverseModeVisitor
       : public clang::ConstStmtVisitor<ReverseModeVisitor, StmtDiff>,
         public VisitorBase {
-  
-  private:
+  protected:
     // FIXME: We should remove friend-dependency of the plugin classes here.
     // For this we will need to separate out AST related functions in
     // a separate namespace, as well as add getters/setters function of
@@ -42,6 +48,8 @@ namespace clad {
     /// the reverse mode we also accumulate Stmts for the reverse pass which
     /// will be executed on return.
     std::vector<Stmts> m_Reverse;
+    /// Storing expressions to delete/free memory in the reverse pass.
+    Stmts m_DeallocExprs;
     /// Stack is used to pass the arguments (dfdx) to further nodes
     /// in the Visit method.
     std::stack<clang::Expr*> m_Stack;
@@ -49,42 +57,46 @@ namespace clad {
     /// that will be put immediately in the beginning of derivative function
     /// block.
     Stmts m_Globals;
-    //// A reference to the output parameter of the gradient function.
-    clang::Expr* m_Result;
+    /// Global GPU args of the function.
+    std::unordered_set<const clang::ParmVarDecl*> m_CUDAGlobalArgs;
     /// A flag indicating if the Stmt we are currently visiting is inside loop.
     bool isInsideLoop = false;
     /// Output variable of vector-valued function
     std::string outputArrayStr;
-    /// Stores the pop index values for arrays in reverse mode.This is required
-    /// to maintain the correct statement order when the current block has
-    /// delayed emission i.e. assignment LHS.
-    Stmts m_PopIdxValues;
     std::vector<Stmts> m_LoopBlock;
+    /// This expression checks if the forward pass loop was terminted due to
+    /// break. It is used to determine whether to run the loop cond
+    /// differentiation. One additional time.
+    clang::Expr* m_CurrentBreakFlagExpr;
+
     unsigned outputArrayCursor = 0;
     unsigned numParams = 0;
-    bool isVectorValued = false;
-    bool use_enzyme = false;
     // FIXME: Should we make this an object instead of a pointer?
     // Downside of making it an object: We will need to include
     // 'MultiplexExternalRMVSource.h' file
     MultiplexExternalRMVSource* m_ExternalSource = nullptr;
     clang::Expr* m_Pullback = nullptr;
     const char* funcPostfix() const {
-      if (isVectorValued)
+      if (m_DiffReq.Mode == DiffMode::jacobian)
         return "_jac";
-      else if (use_enzyme)
+      if (m_DiffReq.use_enzyme)
         return "_grad_enzyme";
-      else
-        return "_grad";
+      return "_grad";
     }
 
     /// Removes the local const qualifiers from a QualType and returns a new
     /// type.
     static clang::QualType
     getNonConstType(clang::QualType T, clang::ASTContext& C, clang::Sema& S) {
-        clang::Qualifiers quals(T.getQualifiers());
-        quals.removeConst();
-        return S.BuildQualifiedType(T.getUnqualifiedType(), noLoc, quals);
+      bool isLValueRefType = T->isLValueReferenceType();
+      T = T.getNonReferenceType();
+      clang::Qualifiers quals(T.getQualifiers());
+      quals.removeConst();
+      clang::QualType nonConstType =
+          S.BuildQualifiedType(T.getUnqualifiedType(), noLoc, quals);
+      if (isLValueRefType)
+        return C.getLValueReferenceType(nonConstType);
+      return nonConstType;
     }
     // Function to Differentiate with Clad as Backend
     void DifferentiateWithClad();
@@ -92,6 +104,11 @@ namespace clad {
     // Function to Differentiate with Enzyme as Backend
     void DifferentiateWithEnzyme();
 
+    /// Tries to find and build call to user-provided `_forw` function.
+    clang::Expr* BuildCallToCustomForwPassFn(
+        const clang::Expr* callSite, llvm::ArrayRef<clang::Expr*> primalArgs,
+        llvm::ArrayRef<clang::Expr*> derivedArgs, clang::Expr* baseExpr);
+
   public:
     using direction = rmv::direction;
     clang::Expr* dfdx() {
@@ -111,28 +128,6 @@ namespace clad {
       return result;
     }
 
-    /// This visit method explicitly sets `dfdx` to `nullptr` for this visit.
-    ///
-    /// This method is helpful when we need derivative of some expression but we
-    /// do not want `_d_expression += dfdx` statments to be (automatically)
-    /// added.
-    ///
-    /// FIXME: Think of a better way for handling this situation. Maybe we
-    /// should improve the overall dfdx design and approach. One other way of
-    /// designing `VisitWithExplicitNoDfDx` in a more general way is
-    /// to develop a function that takes an expression E and returns the
-    /// corresponding derivative without any side effects. The difference
-    /// between this function and the current `VisitWithExplicitNoDfDx` will be
-    /// 1) better intent through the function name 2) We will also get
-    /// derivatives of expressions other than `DeclRefExpr` and `MemberExpr`.
-    StmtDiff VisitWithExplicitNoDfDx(const clang::Stmt* stmt) {
-      m_Stack.push(nullptr);
-      auto result =
-          clang::ConstStmtVisitor<ReverseModeVisitor, StmtDiff>::Visit(stmt);
-      m_Stack.pop();
-      return result;
-    }
-
     /// Get the latest block of code (i.e. place for statements output).
     Stmts& getCurrentBlock(direction d = direction::forward) {
       if (d == direction::forward)
@@ -143,24 +138,33 @@ namespace clad {
     /// Create new block.
     Stmts& beginBlock(direction d = direction::forward) {
       if (d == direction::forward)
-        m_Blocks.push_back({});
+        m_Blocks.emplace_back();
       else
-        m_Reverse.push_back({});
+        m_Reverse.emplace_back();
       return getCurrentBlock(d);
     }
     /// Remove the block from the stack, wrap it in CompoundStmt and return it.
     clang::CompoundStmt* endBlock(direction d = direction::forward) {
       if (d == direction::forward) {
-        auto CS = MakeCompoundStmt(getCurrentBlock(direction::forward));
+        auto* CS = MakeCompoundStmt(getCurrentBlock(direction::forward));
         m_Blocks.pop_back();
         return CS;
       } else {
-        auto CS = MakeCompoundStmt(getCurrentBlock(direction::reverse));
+        auto* CS = MakeCompoundStmt(getCurrentBlock(direction::reverse));
         std::reverse(CS->body_begin(), CS->body_end());
         m_Reverse.pop_back();
         return CS;
       }
     }
+
+    Stmts EndBlockWithoutCreatingCS(direction d = direction::forward) {
+      auto blk = getCurrentBlock(d);
+      if (d == direction::forward)
+        m_Blocks.pop_back();
+      else
+        m_Reverse.pop_back();
+      return blk;
+    }
     /// Output a statement to the current block. If Stmt is null or is an unused
     /// expression, it is not output and false is returned.
     bool addToCurrentBlock(clang::Stmt* S, direction d = direction::forward) {
@@ -174,6 +178,14 @@ namespace clad {
     /// \returns True if the statement was added to the block, false otherwise.
     bool AddToGlobalBlock(clang::Stmt* S) { return addToBlock(S, m_Globals); }
 
+    /// Updates size references in VariableArrayType and replaces
+    /// std::initializer_list with clad::array.
+    clang::QualType CloneType(clang::QualType T);
+
+    /// If E is a CXXSTDInializerListExpr, returns its size expr.
+    /// Otherwise, returns nullptr;
+    clang::Expr* getStdInitListSizeExpr(const clang::Expr* E);
+
     /// Stores the result of an expression in a temporary variable (of the same
     /// type as is the result of the expression) and returns a reference to it.
     /// If force decl creation is true, this will allways create a temporary
@@ -182,26 +194,22 @@ namespace clad {
     /// direct references in intermediate variables)
     clang::Expr* StoreAndRef(clang::Expr* E, direction d = direction::forward,
                              llvm::StringRef prefix = "_t",
-                             bool forceDeclCreation = false,
-                             clang::VarDecl::InitializationStyle IS =
-                                 clang::VarDecl::InitializationStyle::CInit) {
+                             bool forceDeclCreation = false) {
       assert(E && "cannot infer type from null expression");
       return StoreAndRef(E, getNonConstType(E->getType(), m_Context, m_Sema), d,
-                         prefix, forceDeclCreation, IS);
+                         prefix, forceDeclCreation);
     }
 
     /// An overload allowing to specify the type for the variable.
     clang::Expr* StoreAndRef(clang::Expr* E, clang::QualType Type,
                              direction d = direction::forward,
                              llvm::StringRef prefix = "_t",
-                             bool forceDeclCreation = false,
-                             clang::VarDecl::InitializationStyle IS =
-                                 clang::VarDecl::InitializationStyle::CInit) {
+                             bool forceDeclCreation = false) {
       // Name reverse temporaries as "_r" instead of "_t".
       if ((d == direction::reverse) && (prefix == "_t"))
         prefix = "_r";
       return VisitorBase::StoreAndRef(E, Type, getCurrentBlock(d), prefix,
-                                      forceDeclCreation, IS);
+                                      forceDeclCreation);
     }
 
     /// For an expr E, decides if it is useful to store it in a global temporary
@@ -209,6 +217,10 @@ namespace clad {
     /// to avoid recomputiation.
     bool UsefulToStoreGlobal(clang::Expr* E);
 
+    /// For an expr E, decides if we should recompute it or store it.
+    /// This is the central point for checkpointing.
+    bool ShouldRecompute(const clang::Expr* E);
+
     /// Builds a variable declaration and stores it in the function
     /// global scope.
     ///
@@ -227,17 +239,18 @@ namespace clad {
     /// it into m_Globals block (to be inserted into the beginning of fn's
     /// body). Returns reference R to the created declaration. If E is not null,
     /// puts an additional assignment statement (R = E) in the forward block.
-    /// Alternatively, if isInsideLoop is true, stores E in a stack. Returns
-    /// StmtDiff, where .getExpr() is intended to be used in forward pass and
-    /// .getExpr_dx() in the reverse pass. Two expressions can be different in
-    /// some cases, e.g. clad::push/pop inside loops.
-    StmtDiff GlobalStoreAndRef(clang::Expr* E,
-                               clang::QualType Type,
-                               llvm::StringRef prefix = "_t",
-                               bool force = false);
-    StmtDiff GlobalStoreAndRef(clang::Expr* E,
-                               llvm::StringRef prefix = "_t",
-                               bool force = false);
+    /// Alternatively, if isInsideLoop is true, stores E in a stack S. Puts a
+    /// push statement (clad::push(S, E)) in the forward block and a pop
+    /// statement
+    /// ((clad::pop(S))) in the reverse block. Returns a reference to the top
+    /// of the stack (clad::back(S)).
+    clang::Expr* GlobalStoreAndRef(clang::Expr* E, clang::QualType Type,
+                                   llvm::StringRef prefix = "_t",
+                                   bool force = false);
+    clang::Expr* GlobalStoreAndRef(clang::Expr* E,
+                                   llvm::StringRef prefix = "_t",
+                                   bool force = false);
+    StmtDiff StoreAndRestore(clang::Expr* E, llvm::StringRef prefix = "_t");
 
     //// A type returned by DelayedGlobalStoreAndRef
     /// .Result is a reference to the created (yet uninitialized) global
@@ -249,8 +262,19 @@ namespace clad {
     struct DelayedStoreResult {
       ReverseModeVisitor& V;
       StmtDiff Result;
+      clang::VarDecl* Declaration;
       bool isConstant;
       bool isInsideLoop;
+      bool isFnScope;
+      bool needsUpdate;
+      clang::Expr* Placeholder;
+      DelayedStoreResult(ReverseModeVisitor& pV, StmtDiff pResult,
+                         clang::VarDecl* pDeclaration, bool pIsInsideLoop,
+                         bool pIsFnScope, bool pNeedsUpdate = false,
+                         clang::Expr* pPlaceholder = nullptr)
+          : V(pV), Result(pResult), Declaration(pDeclaration),
+            isInsideLoop(pIsInsideLoop), isFnScope(pIsFnScope),
+            needsUpdate(pNeedsUpdate), Placeholder(pPlaceholder) {}
       void Finalize(clang::Expr* New);
     };
 
@@ -263,7 +287,8 @@ namespace clad {
     /// This is what DelayedGlobalStoreAndRef does. E is expected to be the
     /// original (uncloned) expression.
     DelayedStoreResult DelayedGlobalStoreAndRef(clang::Expr* E,
-                                                llvm::StringRef prefix = "_t");
+                                                llvm::StringRef prefix = "_t",
+                                                bool forceStore = false);
 
     struct CladTapeResult {
       ReverseModeVisitor& V;
@@ -290,14 +315,35 @@ namespace clad {
     CladTapeResult MakeCladTapeFor(clang::Expr* E,
                                    llvm::StringRef prefix = "_t");
 
+    /// A function to get the multi-argument "central_difference"
+    /// call expression for the given arguments.
+    ///
+    /// \param[in] targetFuncCall The function to get the derivative for.
+    /// \param[in] retType The return type of the target call expression.
+    /// \param[in] dfdx The dfdx corresponding to this call expression.
+    /// \param[in] numArgs The total number of 'args'.
+    /// \param[in] PreCallStmts The built statements to add to block
+    /// before the call to the derived function.
+    /// \param[in] PostCallStmts The built statements to add to block
+    /// after the call to the derived function.
+    /// \param[in] args All the arguments to the target function.
+    /// \param[in] outputArgs The output gradient arguments.
+    ///
+    /// \returns The derivative function call.
+    clang::Expr* GetMultiArgCentralDiffCall(
+        clang::Expr* targetFuncCall, clang::QualType retType, unsigned numArgs,
+        clang::Expr* dfdx, llvm::SmallVectorImpl<clang::Stmt*>& PreCallStmts,
+        llvm::SmallVectorImpl<clang::Stmt*>& PostCallStmts,
+        llvm::SmallVectorImpl<clang::Expr*>& args,
+        llvm::SmallVectorImpl<clang::Expr*>& outputArgs,
+        clang::Expr* CUDAExecConfig = nullptr);
+
   public:
-    ReverseModeVisitor(DerivativeBuilder& builder);
-    ~ReverseModeVisitor();
+    ReverseModeVisitor(DerivativeBuilder& builder, const DiffRequest& request);
+    virtual ~ReverseModeVisitor();
 
     ///\brief Produces the gradient of a given function.
     ///
-    ///\param[in] FD - the function that will be differentiated.
-    ///
     ///\returns The gradient of the function, potentially created enclosing
     /// context and if generated, its overload.
     ///
@@ -314,42 +360,90 @@ namespace clad {
     /// Improved naming scheme is required. Hence, we append the indices to of
     /// the requested parameters to 'f_grad', i.e. in the previous example "x,
     /// y" will give 'f_grad_0_1' and "x, z" will give 'f_grad_0_2'.
-    DerivativeAndOverload Derive(const clang::FunctionDecl* FD,
-                                 const DiffRequest& request);
-    DerivativeAndOverload DerivePullback(const clang::FunctionDecl* FD,
-                                         const DiffRequest& request);
+    DerivativeAndOverload Derive();
     StmtDiff VisitArraySubscriptExpr(const clang::ArraySubscriptExpr* ASE);
     StmtDiff VisitBinaryOperator(const clang::BinaryOperator* BinOp);
     StmtDiff VisitCallExpr(const clang::CallExpr* CE);
-    StmtDiff VisitCompoundStmt(const clang::CompoundStmt* CS);
+    virtual StmtDiff VisitCompoundStmt(const clang::CompoundStmt* CS);
     StmtDiff VisitConditionalOperator(const clang::ConditionalOperator* CO);
     StmtDiff VisitCXXBoolLiteralExpr(const clang::CXXBoolLiteralExpr* BL);
+    StmtDiff VisitCharacterLiteral(const clang::CharacterLiteral* CL);
+    StmtDiff VisitStringLiteral(const clang::StringLiteral* SL);
     StmtDiff VisitCXXDefaultArgExpr(const clang::CXXDefaultArgExpr* DE);
-    StmtDiff VisitDeclRefExpr(const clang::DeclRefExpr* DRE);
+    StmtDiff
+    VisitCXXFunctionalCastExpr(const clang::CXXFunctionalCastExpr* FCE);
+    virtual StmtDiff VisitDeclRefExpr(const clang::DeclRefExpr* DRE);
     StmtDiff VisitDeclStmt(const clang::DeclStmt* DS);
     StmtDiff VisitFloatingLiteral(const clang::FloatingLiteral* FL);
+    StmtDiff VisitCXXForRangeStmt(const clang::CXXForRangeStmt* FRS);
     StmtDiff VisitForStmt(const clang::ForStmt* FS);
     StmtDiff VisitIfStmt(const clang::IfStmt* If);
     StmtDiff VisitImplicitCastExpr(const clang::ImplicitCastExpr* ICE);
+    StmtDiff
+    VisitImplicitValueInitExpr(const clang::ImplicitValueInitExpr* IVIE);
+    StmtDiff VisitCStyleCastExpr(const clang::CStyleCastExpr* CSCE);
+    StmtDiff VisitPseudoObjectExpr(const clang::PseudoObjectExpr* POE);
     StmtDiff VisitInitListExpr(const clang::InitListExpr* ILE);
     StmtDiff VisitIntegerLiteral(const clang::IntegerLiteral* IL);
     StmtDiff VisitMemberExpr(const clang::MemberExpr* ME);
     StmtDiff VisitParenExpr(const clang::ParenExpr* PE);
-    StmtDiff VisitReturnStmt(const clang::ReturnStmt* RS);
+    virtual StmtDiff VisitReturnStmt(const clang::ReturnStmt* RS);
     StmtDiff VisitStmt(const clang::Stmt* S);
-    StmtDiff VisitUnaryOperator(const clang::UnaryOperator* UnOp);
+    virtual StmtDiff VisitUnaryOperator(const clang::UnaryOperator* UnOp);
+    StmtDiff
+    VisitUnaryExprOrTypeTraitExpr(const clang::UnaryExprOrTypeTraitExpr* UE);
     StmtDiff VisitExprWithCleanups(const clang::ExprWithCleanups* EWC);
     /// Decl is not Stmt, so it cannot be visited directly.
     StmtDiff VisitWhileStmt(const clang::WhileStmt* WS);
     StmtDiff VisitDoStmt(const clang::DoStmt* DS);
     StmtDiff VisitContinueStmt(const clang::ContinueStmt* CS);
     StmtDiff VisitBreakStmt(const clang::BreakStmt* BS);
+    StmtDiff
+    VisitCXXStdInitializerListExpr(const clang::CXXStdInitializerListExpr* ILE);
     StmtDiff VisitCXXThisExpr(const clang::CXXThisExpr* CTE);
+    StmtDiff VisitCXXNewExpr(const clang::CXXNewExpr* CNE);
+    StmtDiff VisitCXXDeleteExpr(const clang::CXXDeleteExpr* CDE);
     StmtDiff VisitCXXConstructExpr(const clang::CXXConstructExpr* CE);
     StmtDiff
     VisitMaterializeTemporaryExpr(const clang::MaterializeTemporaryExpr* MTE);
     StmtDiff VisitCXXStaticCastExpr(const clang::CXXStaticCastExpr* SCE);
-    VarDeclDiff DifferentiateVarDecl(const clang::VarDecl* VD);
+    StmtDiff VisitCXXConstCastExpr(const clang::CXXConstCastExpr* CCE);
+    StmtDiff VisitSwitchStmt(const clang::SwitchStmt* SS);
+    StmtDiff VisitCaseStmt(const clang::CaseStmt* CS);
+    StmtDiff VisitDefaultStmt(const clang::DefaultStmt* DS);
+    DeclDiff<clang::VarDecl> DifferentiateVarDecl(const clang::VarDecl* VD,
+                                                  bool keepLocal = false);
+    StmtDiff VisitSubstNonTypeTemplateParmExpr(
+        const clang::SubstNonTypeTemplateParmExpr* NTTP);
+    StmtDiff
+    VisitCXXNullPtrLiteralExpr(const clang::CXXNullPtrLiteralExpr* NPE);
+    StmtDiff VisitNullStmt(const clang::NullStmt* NS) {
+      return StmtDiff{Clone(NS), Clone(NS)};
+    };
+
+    /// Helper function that checks whether the function to be derived
+    /// is meant to be executed only by the GPU
+    bool shouldUseCudaAtomicOps(const clang::Expr* E);
+
+    /// Add call to cuda::atomicAdd for the given LHS and RHS expressions.
+    ///
+    /// \param[in] LHS The left-hand side expression.
+    ///
+    /// \param[in] RHS The right-hand side expression.
+    ///
+    /// \returns The atomicAdd call expression.
+    clang::Expr* BuildCallToCudaAtomicAdd(clang::Expr* LHS, clang::Expr* RHS);
+
+    /// Check whether this is an assignment to a malloc or a realloc call for a
+    /// derivative variable and build a call to memset to follow the memory
+    /// allocation in order to properly intialize the memory to zero. \param[in]
+    /// LHS The left-hand side expression of the assignment. \param[in] RHS The
+    /// right-hand side expression of the assignment.
+    /// @returns The call to memset if the condition is met, otherwise nullptr.
+    clang::Expr* CheckAndBuildCallToMemset(clang::Expr* LHS, clang::Expr* RHS);
+
+    static DeclDiff<clang::StaticAssertDecl>
+    DifferentiateStaticAssertDecl(const clang::StaticAssertDecl* SAD);
 
     /// A helper method to differentiate a single Stmt in the reverse mode.
     /// Internally, calls Visit(S, expr). Its result is wrapped into a
@@ -377,24 +471,17 @@ namespace clad {
            "attempt to differentiate unsupported operator, ignored.",
            args);
     }
-    /// Builds an overload for the gradient function that has derived params for
-    /// all the arguments of the requested function and it calls the original
-    /// gradient function internally
-    clang::FunctionDecl* CreateGradientOverload();
 
     /// Returns the type that should be used to represent the derivative of a
-    /// variable of type `yType` with respect to a parameter variable of type
-    /// `xType`.
     ///
     /// FIXME: Parameter derivative type rules are different from the derivative
     /// type rules for local variables. We should remove this inconsistency.
     /// See the following issue for more details:
     /// https://github.com/vgvassilev/clad/issues/385
-    clang::QualType GetParameterDerivativeType(clang::QualType yType,
-                                               clang::QualType xType);
+    clang::QualType GetParameterDerivativeType(clang::QualType Type);
 
     /// Allows to easily create and manage a counter for counting the number of
-    /// executed iterations of a loop. 
+    /// executed iterations of a loop.
     ///
     /// It is required to save the number of executed iterations to use the
     /// same number of iterations in the reverse pass.
@@ -406,6 +493,7 @@ namespace clad {
       clang::Expr *m_Pop = nullptr;
       clang::Expr *m_Push = nullptr;
       ReverseModeVisitor& m_RMV;
+      clang::VarDecl* m_numRevIterations = nullptr;
 
     public:
       LoopCounter(ReverseModeVisitor& RMV);
@@ -413,11 +501,11 @@ namespace clad {
       /// for counter; otherwise, returns nullptr.
       clang::Expr* getPush() const { return m_Push; }
 
-      /// Returns `clad::pop(_t)` expression if clad tape is used for 
+      /// Returns `clad::pop(_t)` expression if clad tape is used for
       /// for counter; otherwise, returns nullptr.
       clang::Expr* getPop() const { return m_Pop; }
 
-      /// Returns reference to the last object of the clad tape if clad tape 
+      /// Returns reference to the last object of the clad tape if clad tape
       /// is used as the counter; otherwise returns reference to the counter
       /// variable.
       clang::Expr* getRef() const { return m_Ref; }
@@ -434,9 +522,18 @@ namespace clad {
 
       /// Returns `ConditionResult` object for the counter.
       clang::Sema::ConditionResult getCounterConditionResult() {
-        return m_RMV.m_Sema.ActOnCondition(m_RMV.m_CurScope, noLoc, m_Ref,
+        return m_RMV.m_Sema.ActOnCondition(m_RMV.getCurrentScope(), noLoc,
+                                           m_Ref,
                                            clang::Sema::ConditionKind::Boolean);
       }
+
+      /// Sets the number of reverse iterations to be executed.
+      void setNumRevIterations(clang::VarDecl* numRevIterations) {
+        m_numRevIterations = numRevIterations;
+      }
+
+      /// Returns the number of reverse iterations to be executed.
+      clang::VarDecl* getNumRevIterations() const { return m_numRevIterations; }
     };
 
     /// Helper function to differentiate a loop body.
@@ -457,13 +554,13 @@ namespace clad {
                                    clang::Stmt* forLoopIncDiff = nullptr,
                                    bool isForLoop = false);
 
-    /// This class modifies forward and reverse blocks of the loop
+    /// This class modifies forward and reverse blocks of the loop/switch
     /// body so that `break` and `continue` statements are correctly
-    /// handled. `break` and `continue` statements are handled by 
+    /// handled. `break` and `continue` statements are handled by
     /// enclosing entire reverse block loop body in a switch statement
     /// and only executing the statements, with the help of case labels,
-    /// that were executed in the associated forward iteration. This is 
-    /// determined by keeping track of which `break`/`continue` statement 
+    /// that were executed in the associated forward iteration. This is
+    /// determined by keeping track of which `break`/`continue` statement
     /// was hit in which iteration and that in turn helps to determine which
     /// case label should be selected.
     ///
@@ -491,7 +588,7 @@ namespace clad {
       /// \note `m_ControlFlowTape` is only initialized if the body contains
       /// `continue` or `break` statement.
       std::unique_ptr<CladTapeResult> m_ControlFlowTape;
-      
+
       /// Each `break` and `continue` statement is assigned a unique number,
       /// starting from 1, that is used as the case label corresponding to that `break`/`continue`
       /// statement. `m_CaseCounter` stores the value that was used for last
@@ -514,7 +611,10 @@ namespace clad {
       clang::Expr* CreateCFTapePushExpr(std::size_t value);
 
     public:
-      BreakContStmtHandler(ReverseModeVisitor& RMV) : m_RMV(RMV) {}
+      bool m_IsInvokedBySwitchStmt = false;
+
+      BreakContStmtHandler(ReverseModeVisitor& RMV, bool forSwitchStmt = false)
+          : m_RMV(RMV), m_IsInvokedBySwitchStmt(forSwitchStmt) {}
 
       /// Begins control flow switch statement scope.
       /// Control flow switch statement is used to refer to the
@@ -530,11 +630,16 @@ namespace clad {
       /// control flow switch statement.
       clang::CaseStmt* GetNextCFCaseStmt();
 
-      /// Builds and returns `clad::push(TapeRef, m_CurrentCounter)` 
+      /// Builds and returns `clad::push(TapeRef, m_CurrentCounter)`
       /// expression, where `TapeRef` and `m_CurrentCounter` are replaced
       /// by their actual values respectively.
       clang::Stmt* CreateCFTapePushExprToCurrentCase();
 
+      /// Builds and return `clad::back(TapeRef) != m_CaseCounter`
+      /// expression, where `TapeRef` and `m_CaseCounter` are replaced
+      /// by their actual values respectively
+      clang::Expr* CreateCFTapeBackExprForCurrentCase();
+
       /// Does final modifications on forward and reverse blocks
       /// so that `break` and `continue` statements are handled
       /// accurately.
@@ -546,14 +651,14 @@ namespace clad {
     BreakContStmtHandler* GetActiveBreakContStmtHandler() {
       return &m_BreakContStmtHandlers.back();
     }
-    BreakContStmtHandler* PushBreakContStmtHandler() {
-      m_BreakContStmtHandlers.emplace_back(*this);
+    BreakContStmtHandler* PushBreakContStmtHandler(bool forSwitchStmt = false) {
+      m_BreakContStmtHandlers.emplace_back(*this, forSwitchStmt);
       return &m_BreakContStmtHandlers.back();
     }
     void PopBreakContStmtHandler() {
       m_BreakContStmtHandlers.pop_back();
     }
-                                   
+
     /// Registers an external RMV source.
     ///
     /// Multiple external RMV source can be registered by calling this function
@@ -561,26 +666,62 @@ namespace clad {
     ///\paramp[in] source An external RMV source
     void AddExternalSource(ExternalRMVSource& source);
 
-    /// Computes and returns the sequence of derived function parameter types.
-    ///
-    /// Information about the original function and the differentiation mode
-    /// are taken from the data member variables. In particular, `m_Function`,
-    /// `m_Mode` data members should be correctly set before using this
-    /// function.
-    llvm::SmallVector<clang::QualType, 8> ComputeParamTypes(const DiffParams& diffParams);
+    /// Computes and returns the derived function prototype.
+    clang::QualType ComputeDerivativeFunctionType();
 
     /// Builds and returns the sequence of derived function parameters.
-    ///
-    /// Information about the original function, derived function, derived
-    /// function parameter types and the differentiation mode are implicitly
-    /// taken from the data member variables. In particular, `m_Function`,
-    /// `m_Mode` and `m_Derivative` should be correctly set before using this
-    /// function.
-    llvm::SmallVector<clang::ParmVarDecl*, 8>
-    BuildParams(DiffParams& diffParams);
+    void BuildParams(llvm::SmallVectorImpl<clang::ParmVarDecl*>& params);
 
     clang::QualType ComputeAdjointType(clang::QualType T);
-    clang::QualType ComputeParamType(clang::QualType T);
+    /// Stores data required for differentiating a switch statement.
+    struct SwitchStmtInfo {
+      llvm::SmallVector<clang::SwitchCase*, 16> cases;
+      clang::Expr* switchStmtCond = nullptr;
+      clang::IfStmt* defaultIfBreakExpr = nullptr;
+    };
+
+    /// Maintains a stack of `SwitchStmtInfo`.
+    llvm::SmallVector<SwitchStmtInfo, 4> m_SwitchStmtsData;
+
+    SwitchStmtInfo* GetActiveSwitchStmtInfo() {
+      return &m_SwitchStmtsData.back();
+    }
+
+    SwitchStmtInfo* PushSwitchStmtInfo() {
+      m_SwitchStmtsData.emplace_back();
+      return &m_SwitchStmtsData.back();
+    }
+
+    void PopSwitchStmtInfo() { m_SwitchStmtsData.pop_back(); }
+
+    struct ConstructorPullbackCallInfo {
+      clang::CallExpr* pullbackCE = nullptr;
+      size_t thisAdjointArgIdx = std::numeric_limits<size_t>::max();
+      void updateThisParmArgs(clang::Expr* thisE, clang::Expr* dThisE) const;
+      ConstructorPullbackCallInfo() = default;
+      ConstructorPullbackCallInfo(clang::CallExpr* pPullbackCE,
+                                  size_t pThisAdjointArgIdx)
+          : pullbackCE(pPullbackCE), thisAdjointArgIdx(pThisAdjointArgIdx) {}
+
+      bool empty() const { return !pullbackCE; }
+    };
+
+    void setConstructorPullbackCallInfo(clang::CallExpr* pullbackCE,
+                                        size_t thisAdjointArgIdx) {
+      m_ConstructorPullbackCallInfo = {pullbackCE, thisAdjointArgIdx};
+    }
+
+    ConstructorPullbackCallInfo getConstructorPullbackCallInfo() {
+      return m_ConstructorPullbackCallInfo;
+    }
+
+    void resetConstructorPullbackCallInfo() {
+      m_ConstructorPullbackCallInfo = ConstructorPullbackCallInfo{};
+    }
+
+  private:
+    ConstructorPullbackCallInfo m_ConstructorPullbackCallInfo;
+    bool m_TrackConstructorPullbackInfo = false;
   };
 } // end namespace clad
 
