@@ -52,7 +52,7 @@
 #define R__HAS_DATAFRAME /**/
 #define R__HAS_ROOT7 /**/
 #undef R__LESS_INCLUDES /**/
-#undef R__HAS_TBB /**/
+#define R__HAS_TBB /**/
 
 #if defined(R__HAS_VECCORE) && defined(R__HAS_VC)
 #ifndef VECCORE_ENABLE_VC
@@ -64,7 +64,7 @@
 #define R__HAS_DEFAULT_ZLIB  /**/
 #undef R__HAS_DEFAULT_LZMA  /**/
 #undef R__HAS_DEFAULT_ZSTD  /**/
-#define R__HAS_CLOUDFLARE_ZLIB /**/
+#undef R__HAS_CLOUDFLARE_ZLIB /**/
 
 #define R__HAS_TMVACPU /**/
 #undef R__HAS_TMVAGPU /**/
